/* tracedsubs.c
  traced is a network daemon
   - when it starts up, it reads 1 file--traced.conf:
	containing connections to be traced 
	web100 variables to be logged from those connections
   - the daemon alternates checking for connections and sleeping for 30 seconds
   - in tracing it:
	reads the /proc/web100 directory and keeps a linked list of all active
	  connections
	checks all web100 connections it finds against the ones to be logged and
	  if there is a match--creates a thread to monitor that flow 
	    and periodically log values every 80 milliseconds (RTT) to a 
	    cid.trace file 
	cleans up the linked list of connections--removing any 
         connections that were not found in the /proc/web100 directory
	 on the last read
*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <stdlib.h>
#include <pthread.h>
#include "web100.h"
#include "web100-int.h"


#define ERRLOG "traced.errlog"
#define CONFIGFILE "traced.conf"

/*cids with these ports do not need to be logged*/
#define SSH_PORT 22
#define HTTP_PORT 80
#define BWSERVER 7123
#define LBLPORT1 5002
#define LBLPORT2 8002
#define GRID_CTL 2811

/*Connection Stage */
#define NOINTEREST 0
#define MONITOR    1
#define LOGGED     2

/*GLOBAL STRUCTURES AND VARIABLES*/
struct web100_agent *agent;

struct sysvar {
  struct web100_var *varptr;
  int mode;
  int length;
};
struct sysvar *sysvars;

struct active_cid {
  int cid;
  int stage;
  int up;
  struct web100_connection *connsave;
  struct active_cid *nextconn;
};

struct active_cid *CidRoot=NULL;
struct web100_var *CidState;
struct web100_var *Duration;
struct web100_var *StartTimeUsec;
struct web100_var *ThruBytesAcked;
struct web100_var *RemPort;
struct web100_var *RemAddress;
struct web100_var *LocalPort;
struct web100_var *LocalAddress;

struct web100_connection_spec *ConfigConnects;
int Connections, VarCount;
FILE *ef;

int traced_init();
void read_config();
void WalkProcWeb100();
struct active_cid *AddCid(struct web100_connection *);
struct active_cid *FindCid(int);
struct web100_connection *web100_ConnectionStart(struct web100_agent *, char *);
void *MonitorConn(void *);
double seconds();

int err_sys(char *msg) {
  ef = fopen(ERRLOG, "a");
  if(ef==NULL) {
    fprintf(stderr, "%s\n", "err_sys: traced.errlog open Error");
    exit(1);
  }
  fprintf(ef, "%s\n", msg);
  fclose(ef);
}

/*set up the necessary structures for monitoring connections at the
    beginning
*/
int traced_init() {
  struct web100_group *group;
  int i, retcode;

  if((agent=web100_attach(WEB100_AGENT_TYPE_LOCAL, NULL)) == NULL) {
    fprintf(stderr, "%s\n", "traced_init: ERROR No agent found!");
    exit(2);
  }
  read_config(); 

  if((group = web100_group_find(agent, "read"))==NULL) {
    fprintf(stderr, "_init: can't locate /read\n");
    exit(1);
  }
  if((CidState = web100_var_find(group, "State"))==NULL) {
    fprintf(stderr, "_init: can't locate CidSstate\n");
    exit(1);
  }
  if((Duration = web100_var_find(group, "Duration"))==NULL) {
    fprintf(stderr, "_init: can't locate Duration\n");
    exit(1);
  }
  if((StartTimeUsec = web100_var_find(group, "StartTimeUsec"))==NULL){
    fprintf(stderr, "_init: can't locate StartTimeUsec\n");
    exit(1);
  }
  if((ThruBytesAcked = web100_var_find(group, "ThruBytesAcked"))==NULL){
    fprintf(stderr, "_init: can't locate ThruBytesAcked\n");
    exit(1);
  }
  if((RemPort = web100_var_find(group, "RemPort"))==NULL) {
    fprintf(stderr, "_init: can't locate RemPort\n");
    exit(1);
  }
  if((RemAddress = web100_var_find(group, "RemAddress"))==NULL) {
    fprintf(stderr, "_init: can't locate RemAddress\n");
    exit(1);
  }
  if((LocalPort = web100_var_find(group, "LocalPort"))==NULL) {
    fprintf(stderr, "_init: can't locate LocalPort\n");
    exit(1);
  }
  if((LocalAddress = web100_var_find(group, "LocalAddress"))==NULL) {
    fprintf(stderr, "_init: can't locate LocalAddress\n");
    exit(1);
  }
  return(0);
}

/*load the connections from the traced.conf file into a table so statistics
    from these connections can be logged
*/
void read_config() {
  FILE *cf;
  char laddr[20], raddr[20], line[80], VarName[32], mode[5], *ptr;
  int lport, rport, i, err=0;
  struct web100_connection_spec *CC;
  struct web100_group *group;
  struct web100_var *varptr;

/* kill -1 pid   will send the process the signal SIGHUP
   to re-read the wad.conf file */
      signal(SIGHUP, read_config);

  if((cf = fopen(CONFIGFILE, "r")) == NULL) {
    fprintf(stderr, "%s\n", "read_config: traced.conf File Open Error");
    exit(1);
  }
  
  free(ConfigConnects);
  while(!(feof(cf) || line[0]=='*')) {
    fgets(line, 79, cf);
    if(line[0]=='#' || line[0]=='*')
      continue;
    Connections++;
  }
  if(Connections<=0) {
    fprintf(stderr, "ERROR: Config File Connections %d!", Connections);
    fclose(cf);
    exit(1);
  }
  while((ptr = fgets(line, 79, cf)) != NULL){
    if(line[0]=='#' || line[0]=='*') {
      fprintf(stderr, "ERROR: too many comments in config file\n");
      exit(1);
    }
    VarCount++;
  }
  if(VarCount<=0) {
    fprintf(stderr, "ERROR: Config File VarCount %d!", VarCount);
    fclose(cf);
    exit(1);
  }
  rewind(cf);

  ConfigConnects = (struct web100_connection_spec *) malloc (Connections*sizeof(struct web100_connection_spec));
  if(ConfigConnects)
    CC = ConfigConnects; 
  else {
    fprintf(stderr, "%s\n", "read_config: ConfigConnects malloc error");
    fclose(cf);
    exit(1);
  }

/*get comment line*/
  fgets(line, 79, cf);

  for(i=0; i<Connections; i++) {
    fscanf(cf, "%s %d %s %d", laddr, &lport, raddr, &rport);
/*convert addresses to unsigned longs and store with ports in table*/
    CC->src_addr = inet_addr((char *)laddr);
    CC->dst_addr = inet_addr((char *)raddr);
    if(CC->dst_addr == 0) {
      fprintf(stderr, "read_config: Remote addresses cannot be 0!\n");
      fclose(cf);
      exit(1);
    }
    CC->src_port = lport;
    CC->dst_port = rport;
    CC++; 
fprintf(stderr, "monitoring %s %d\n", raddr, rport);
  }

/*get \n plus comment line*/
  fgets(line, 79, cf);
  fgets(line, 79, cf);

  free(sysvars);
  sysvars = (struct sysvar *) malloc(VarCount*sizeof(struct sysvar));
  if(sysvars!=NULL) {
    for(i=0; i<VarCount; i++) {
      fscanf(cf, "%s %s", mode, VarName);
      err = web100_agent_find_var_and_group(agent, VarName, &group, &varptr);
      if(err > 0) {
	fprintf(stderr, "err: %d\n", err);
	exit(1);
      }
      sysvars[i].varptr = varptr;
      if((strcmp(mode, "v"))==0)
	sysvars[i].mode = 1;
      else
	sysvars[i].mode = 2;
      sysvars[i].length = varptr->len;
    }
  }
  else {
    fprintf(stderr, "%s\n", "traced_init: Out of Memory!");
    exit(1);
  }
  fprintf(stderr, "Loaded %d trace variables...\n", VarCount);
  fclose(cf);
}

/* WalkProcWeb100() walks the directory tree comparing connections
   found with those in a linked list--those not found are passed to
   TuneConnections() to be added and possibly tuned.  Those found
   in the list are ignored as they have already been dealt with.
   Since this method has no built-in notification for the closing
   of a connection, the state of each connection in the list that is
   being tuned must be checked and, if a connection is in a closed
   state, it is removed from tuning consideration.  Finally, the
   connections are removed from the linked list of active connections
   when it disappears from the directory, /proc/web100.
*/
void WalkProcWeb100() {
  char path[32]="/proc/web100";
  struct web100_connection *conn;
  struct stat statbuf;
  struct dirent *dirp;
  DIR *dp;
  char *ptr;
  unsigned long local, remote;
  int retcode, dead=0, newstage=0, t;
  struct active_cid *cidentry;
  pthread_t mtid;


  if((dp = opendir(path))==NULL) {
    err_sys("WalkProcWeb100: Can't read /proc/web100");
    exit(1);
  }
  ptr = path + strlen(path);
  *ptr++ = '/';
  *ptr = 0;

  while((dirp = readdir(dp)) != NULL) {
    if(strcmp(dirp->d_name, ".")==0 || strcmp(dirp->d_name, "..")==0)
      continue;
    strcpy(ptr, dirp->d_name);
/*If the directory entry is a valid directory */
    if(lstat(path, &statbuf) < 0) {
      fprintf(stderr, "%s %s\n", "lstat error on ", path);
      continue;
    }
    if(S_ISDIR(statbuf.st_mode)) {
	cidentry = FindCid(atoi(dirp->d_name));
	if(cidentry) 
	    continue;
	/*not in active list--need to add*/
/*get addresses/ports for this cid--then add it to linked list*/
        conn = web100_ConnectionStart(agent, dirp->d_name);
	cidentry = AddCid(conn);
        cidentry->stage=FilterByConnection(conn);
/* if stage = MONITOR and connection state > 2 and < 8, create a
   thread to trace variables */
	if(cidentry->stage==MONITOR)  {
fprintf(stderr, "Creating a thread for cid %d: ", cidentry->cid);
          retcode = pthread_create(&mtid, NULL, MonitorConn, (void *)cidentry);
          if(retcode < 0) {
            fprintf(stderr, "WalkProcWeb: pthread_create error\n");
            exit(2);
          }
        }
    }
    *ptr = 0;
  }
  closedir(dp);
  if(CidRoot) 
    dead = CleanUp();
}
struct web100_connection *web100_ConnectionStart(struct web100_agent *agent, char *d_name) {
  web100_connection *cp;
  FILE *fp;
  char filename[64];

  if ((cp = (web100_connection *)malloc(sizeof (web100_connection))) == NULL) {
    perror("connection_start: malloc error");
    return (0);
  }
  cp->cid = atoi(d_name);
  cp->agent = agent;
  cp->logstate = cp->error = 0;
  cp->info.local.next = NULL;
  strcpy(filename, WEB100_ROOT_DIR);
  strcat(filename, d_name);
  strcat(filename, "/read");
  if ((fp = fopen(filename, "r")) == NULL) {
    perror("connection_start: fopen");
    return (0);
  }
  if(fseek(fp, RemPort->offset, SEEK_SET)) {
    printf("RemPort: fseek error!\n");
    exit(2);
  }
  if((fread(&cp->spec.dst_port, RemPort->len, 1, fp)) != 1) {
    printf("RemPort: fread error!\n");
    exit(2);
  }
  if(fseek(fp, RemAddress->offset, SEEK_SET)) {
    printf("RemAddress: fseek error!\n");
    exit(2);
  }
  if (fread(&cp->spec.dst_addr, RemAddress->len, 1, fp) != 1
      || fread(&cp->spec.src_port, LocalPort->len, 1, fp) != 1
      || fread(&cp->spec.src_addr, LocalAddress->len, 1, fp) != 1) {
      perror("fread");
      fprintf(stderr, "ConnectionStart: bad read file format\n");
      return (0);
   }
printf("dst_addr: %u dst_port %u src_addr: %u src_port: %u\n",
cp->spec.dst_addr,cp->spec.dst_port,cp->spec.src_addr,cp->spec.src_port);
   if (fclose(fp))
      perror("connection_start: fclose");
 
  return(cp);

}


/*traverse the list to clear out dead connections*/
int CleanUp() {
  struct active_cid *next, *prev, *tmp;
  int dead=0;

  next = prev = CidRoot;
  while(next) {
    if(next->up == 0) {
      tmp=next;
      if(next==CidRoot)
	CidRoot=next->nextconn;
      prev->nextconn = next = next->nextconn;
      if(tmp->connsave) {
	free(tmp->connsave);
      }
      free(tmp);
      dead++;
    }
    else {
      next->up = 0;
      prev = next;
      next = next->nextconn;
    }
  }
/*if(dead > 0) {
printf("\n");
for(tmp=CidRoot; tmp; tmp=tmp->nextconn)
  printf("%d-->", tmp->cid);
  printf("\n");
}
*/
  return(dead);
}

/*See if a connection is in the list and, if so,  return the address
    and the stage
*/
struct active_cid *FindCid(int cid) {
  struct active_cid *tmp;

  tmp = CidRoot;
  while(tmp) {
    if(tmp->cid == cid) {
      tmp->up = 1;
      return(tmp);
    }
    tmp = tmp->nextconn;
  }
  return(struct active_cid *)(0);
}

/*Add a new connection to a linked list of active connections
*/
struct active_cid *AddCid(struct web100_connection *conn) {
  struct active_cid *new_cid, *next, *prev, *tmp;

  if(((struct active_cid *)new_cid = (struct active_cid *)malloc(sizeof(struct active_cid)))==NULL) {
    err_sys("AddCid: Out of memory\n");
    exit(1);
  }
  new_cid->cid = conn->cid;
  new_cid->stage = 0;
  new_cid->connsave = conn;
  new_cid->up = 1;
  new_cid->nextconn = CidRoot;
  CidRoot = new_cid;
/*
for(tmp=CidRoot; tmp; tmp=tmp->nextconn)
  printf("%d-->", tmp->cid);
printf("\n");
*/
  return(new_cid);
}

/*Filter Out: unwanted ports 
	      connections not included in the config file, traced.conf 
*/
int FilterByConnection(struct web100_connection *conn) {
  struct web100_connection_spec *CC;
  int i=0, aok=0;

  if((conn->spec.dst_port == HTTP_PORT) ||
    (conn->spec.src_port == BWSERVER) ||
    (conn->spec.src_port == LBLPORT1) ||
    (conn->spec.src_port == LBLPORT2) ||
    (conn->spec.dst_port == GRID_CTL) ||
    (conn->spec.dst_port == SSH_PORT) ||
    (conn->spec.src_port == SSH_PORT)) {
    return(NOINTEREST);
  }

  for(i=0, CC=ConfigConnects; i<Connections; i++, CC++) {
    if(conn->spec.dst_addr == CC->dst_addr) {
      if(CC->dst_port > 0)
	if(conn->spec.dst_port == CC->dst_port) {
	  if(aok = CheckLocal(conn, CC))
	    break;
	}
	else
	  continue;
      else {
	if(aok = CheckLocal(conn, CC))
	  break;
      }
    }
  } 
  if(!aok) {
    return(NOINTEREST);
  }
  else 
    return(MONITOR);
}

int CheckLocal(struct web100_connection *conn, struct web100_connection_spec *CC) {

  if(CC->src_addr > 0)
    if(CC->src_addr == conn->spec.src_addr)
      if(CC->src_port > 0)
        if(conn->spec.src_port == CC->src_port)
	  return(1);
        else
	  return(0);
      else
	return(1);
    else
      return(0);
  else
    return(1);
}

/* open the /proc/web100/cid#/read file and retrieve the value for
   State, ThruBytesAcked and other variables specified by the user
   every 1/10 second(100 msec)
   (ONLY from the /proc/web100/read file at present)
*/
void *MonitorConn(void *entry) {
  struct active_cid *cidentry = (struct active_cid *)entry;
  struct web100_connection *conn = cidentry->connsave;
  FILE *sf;
  int fd;
  char filename[32], statfile[32], line[1024], dotsrc[16], *dotdst, buff[80];
  struct web100_snapshot *currsnap, *priorsnap;
  struct web100_group *fp;
  unsigned int val, thrubytessent, databytesdif; 
  unsigned int curstate=0, starttime=0, currtime=0;
  unsigned int run_time, diff_time, run_sec, run_milli;
  time_t stamp=0;
  struct timeval tv;
  int n, i;
  double tempdif, temprun, mbs, run_mbs, stime, etime;
  struct in_addr source, destin;

  pthread_detach(pthread_self());

  fp = CidState->group;
  if((currsnap = web100_snapshot_alloc(fp, conn))==NULL){
    web100_perror("web100_snapshot_alloc");
    return(NULL);
  }
  if(web100_snap(currsnap)) {
    web100_perror("web100_snap");
    fclose(sf);
    return(NULL); 
  }
  if(web100_snap_read( CidState, currsnap, &curstate)) {
    web100_perror("web100_snap_read");
    fclose(sf);
    return(NULL);
  }
  if(web100_snap_read( ThruBytesAcked, currsnap, &thrubytessent)) {
    web100_perror("web100_snap_read");
    fclose(sf);
    return(NULL);
  }

  if((priorsnap = web100_snapshot_alloc(fp, conn))==NULL){
    web100_perror("web100_snapshot_alloc");
    return(NULL);
  }
/*Too Late--connection is closed*/
  if((curstate < 3) || (curstate > 5)) {
fprintf(stderr, "State: %d ThruBytesAcked: %u--bailing out!\n", curstate, thrubytessent);
    web100_snapshot_free(currsnap);
    web100_snapshot_free(priorsnap);
    return(NULL);
  }
  sprintf(statfile, "%d.%s", cidentry->cid, "stats");
  if ((sf = fopen(statfile, "a")) == NULL) {
    perror("MonitorConn: fopen stats");
    return(NULL);
  }
  stamp = time(0);
  source.s_addr = cidentry->connsave->spec.src_addr;
  destin.s_addr = cidentry->connsave->spec.dst_addr;
  strcpy(dotsrc, inet_ntoa(source));
  dotdst = inet_ntoa(destin);
  fprintf(sf, "HDR %s:%u-->%s:%u  %s", 
    dotsrc, cidentry->connsave->spec.src_port, 
    dotdst, cidentry->connsave->spec.dst_port, ctime(&stamp));

fprintf(stderr, "HDR %s:%u-->%s:%u\n", 
    dotsrc, cidentry->connsave->spec.src_port, 
    dotdst, cidentry->connsave->spec.dst_port);

  line[0] = '\0';
  strcat(line, "HDR Current  Cur   Cum ");
  fprintf(sf, "%s\n", line); 
  line[0] = '\0';
  strcat(line, "HDR Time    Mbs   Mbs ");
  for(i=0; i<VarCount; i++) {
    strcat(line, sysvars[i].varptr->name);
    strcat(line, " ");
  }
  fprintf(sf, "%s\n", line); 
  line[0]='\0';

  if(web100_snap_data_copy(priorsnap, currsnap)) {
    web100_perror("web100_snap_data_copy");
    fclose(sf);
    return(NULL);
  }
  if(web100_snap_read( StartTimeUsec, currsnap, &starttime)) {
    web100_perror("web100_snap_read");
    fclose(sf);
    return(NULL);
  }
  if(web100_snap_read( Duration, currsnap, &currtime)) {
    web100_perror("web100_snap_read");
    fclose(sf);
    return(NULL);
  }

  while((curstate > 2) && (curstate < 8)) {

/*do time stuff and figure mbs*/
    if(web100_delta_any(Duration, currsnap, priorsnap, &diff_time))  
          val = -1;
    if(web100_delta_any(ThruBytesAcked, currsnap, priorsnap, &databytesdif))  
          val = -1;
    memcpy(&thrubytessent, currsnap->data+ThruBytesAcked->offset, ThruBytesAcked->len);
    run_time = currtime - starttime;
    temprun = run_time / 1000000.0;
    if(diff_time>0) {
      tempdif = diff_time/1000000.0;
      mbs = ((databytesdif*8) / 1000000.0) / tempdif;
    }
    else
      mbs = ((thrubytessent*8)/1000000.0)/ temprun;
    run_mbs = ((thrubytessent*8)/1000000.0)/ temprun;
    sprintf(buff, "  %6.2f  %6.2f  %3.2f", (currtime/1000000.0), mbs, run_mbs);
    strcat(line, buff);

/* Obtain values or deltas for web100 vars, figure the Mbs, 
   then write this info to a log file */
    for(i=0; i<VarCount; i++) {
      if(sysvars[i].mode == 1) { 
        memcpy(&val, currsnap->data+sysvars[i].varptr->offset, sysvars[i].length);
      }
      else {
        if(web100_delta_any(sysvars[i].varptr, currsnap, priorsnap, &val))  {
	  val = -1; 
        }
      } 
    sprintf(buff, "%10u ", val);
    strcat(line, buff);
    }
    fprintf(sf, "%s\n", line); 
    line[0]='\0';

/*wait one RTT or 1/10 of a second */    
    tv.tv_sec = 0;
    tv.tv_usec = 100000;
    n = select((int)NULL, NULL, NULL, NULL, &tv);
    if(n<0) break;

    if(web100_snap_data_copy(priorsnap, currsnap)) {
      web100_perror("web100_snap_data_copy");
      fclose(sf);
      return(NULL);
    }
    if(web100_snap(currsnap)) {
      if(web100_errno == WEB100_ERR_NOCONNECTION)
        continue;
      web100_perror("web100_snap");
      fclose(sf);
      return(NULL);
    }
    if(web100_snap_read( CidState, currsnap, &curstate)) {
      web100_perror("web100_snap_read");
      fclose(sf);
      return(NULL);
    }
    if(web100_snap_read( Duration, currsnap, &currtime)) {
      web100_perror("web100_snap_read");
      fclose(sf);
      return(NULL);
    }
  }
/*  etime = seconds();
fprintf(sf, "\nseconds(): %f - %f = %f\n", stime, etime, etime - stime);
fprintf(sf, "Duration: %f - %f = %f\n", currtime/1000000.0, starttime/1000000.0, (currtime - starttime)/1000000);
  
  stamp = time(0);
  fprintf(sf, "%s", ctime(&stamp));
*/
  fprintf(sf, "HDR StartTime: %u\n", starttime); 
  if (fclose(sf))
    perror("MonitorConn: fclose");
  web100_snapshot_free(currsnap);
  web100_snapshot_free(priorsnap);
  return(NULL);
}

/**********************************************************************
  Function:     seconds()
  Description:  Calls the function, gettimeofday(), which gets the system's
                notion of the current time expressed in elapsed seconds and
                microseconds since January 1, 1970.
  Returns:      The current time converted to seconds as a double.
**********************************************************************/
double seconds()
{
  #include <sys/time.h>

  struct timeval ru;

  gettimeofday(&ru, (struct timezone *)0);
  return(ru.tv_sec + ((double)ru.tv_usec)/1000000);
}
/*********************************************************************/

